// COOL24+24 (v1.0)
// To the extent possible under law, the author has waived all copyright and
// related or neighboring right to this work (CC0).

// game.js: ゲームのメインの状態遷移を管理
// 定数宣言: 設定項目は https://puyolayersbridge.sakura.ne.jp/games.html 参照
const numSuites = 4;
const numNumbers = 6;
const numCards = numSuites * numNumbers;
const numHands = 4;
const numHandWins = 3;
const cardWinTable = 
  [0, 0, 1, 2, 3, 4, 5, 6, 7, 8,
   10, 12, 14, 16, 18, 20, 25, 30, 35, 40, 45,
   50, 60, 70, 80,
   100, 110, 120, 130, 140, 150, 160, 170, 180, 190,
   200, 220, 240, 260, 280, 300, 350, 400, 450, 500,
   600, 700, 800, 1000];
const handWinName = ["Straight Flush", "Four of a Kind", "Flush"];
const handWinTable = [50, 20, 10];
const handWinMulti = 4;
const suiteNames = ["r", "b", "g", "y"];
const suiteColors = ["#c00", "#00c", "#080", "#880"];
const suiteColorsLite = ["#fcc", "#ccf", "#cfc", "#ff8"];
const numberNames = ["9", "10", "J", "Q", "K", "A"];

const STATE_OVER  = 0;
const STATE_WAIT  = 1;
const STATE_CHECK = 2;
const STATE_DEAL  = 3;
const STATE_MOVE  = 4;

var cardGraphics = new Array;
var timerMain;
var canvas;
var info;
var gameState;
var keyInput;

// ゲームの初期化とセットアップ: 主にカード画像の準備
function initializeGame() {
  for (let i = 0; i < numSuites; i++) {
    for (let j = 0; j < numNumbers; j++) {
      let index = i * numNumbers + j;
      cardGraphics[index] = new CardGraphic(i, j);
    }
  }
  canvas = document.getElementById("cnv");
  info = new GameInfo();
  gameState = STATE_OVER;
  keyInput = "";
  document.onkeypress = (event) => { keyInput = event.key; };
  timerMain = setInterval("gameMain()", 50);
}

// メインループ: 定期的に呼ばれる
function gameMain() {
  ctx = canvas.getContext('2d');
  info.draw(ctx);
  if (gameState == STATE_OVER) {
    // ゲームオーバー（または2周目へ）の表示
    if (keyInput == "Enter") {
      gameState = STATE_CHECK;
      if (info.over == 2) {
        info.resetBoard();
        info.board += 1;
      } else {
        info.resetGame();
      }
    }
  } else if (gameState == STATE_WAIT) {
    // プレイヤーの操作待ち
    let pos = "1234".indexOf(keyInput);
    if (keyInput != "" && pos != -1) {
      if (info.handSelectable(pos)) {
        gameState = STATE_MOVE;
        this.info.handPos[pos] = 11;
      }
    }
  } else if (gameState == STATE_CHECK) {
    // カードを出した場合: 手札配布
    // カードを配った場合: ゲーム終了のチェック
    let c = 0;
    let nc = 0;
    let dealt = false;
    for (let i = 0; i < numHands; i++) {
      if (info.hand[i] == null) {
        // 手札に空きがあり，山札があるなら配る
        if (info.rest.length > 0 && ! dealt) {
          let corr = (info.successive < 20) ? 3.0 - (info.successive * 0.1) : 1.0;
          let sel = info.selectNextCard(corr);
          info.dealNextCard(i, sel);
          dealt = true;
        }
      } else {
        // 手札が出せるかどうかをチェック
        if (info.handSelectable(i)) {
          c += 1;
        } else {
          nc += 1;
        }
      }
    }
    if (c + nc == numHands)
      info.checkHandWin();    // 手札が4枚なら役のチェックをしたあと……
    if (dealt) {
      gameState = STATE_DEAL; // 手札を配ったらアニメーション開始
    } else if (c + nc == 0) {
      info.over = info.board + 2;
      gameState = STATE_OVER; // 手札が空なら勝利
    } else if (c == 0) {
      info.over = 1;
      gameState = STATE_OVER; // 出せないカードしかないなら敗北
    } else {
      gameState = STATE_WAIT; // それ以外はゲーム続行
    }
  } else { // gameState == STATE_DEAL || gameState == STATE_MOVE
    // カードを配る or 場に出すときのアニメーション中
    for (let i = 0; i < numHands; i++) {
      if (info.hand[i] != null && info.handPos[i] != 10) {
        if (info.handPos[i] == 4) {
          info.handPos[i] = 10;
          gameState = STATE_CHECK;
        } else if (info.handPos[i] == 19) {
          info.used.push(info.hand[i]);
          info.top = info.hand[i];
          info.hand[i] = null;
          info.successive += 1;
          gameState = STATE_CHECK;
        } else {
          info.handPos[i] += 1;
        }
      }
    }
  }
  keyInput = "";
}