// COOL24+24 (v1.0)
// To the extent possible under law, the author has waived all copyright and
// related or neighboring right to this work (CC0).

// info.js: ゲームに関する情報保持と画面描画を行うためのクラス
class GameInfo {
  constructor () {
    this.resetGame();
    this.over = 1;
  }

  // ゲーム画面全体を描画する
  draw(ctx) {
    ctx.fillStyle = (this.board == 0) ? "#cff" : "#fcf";
    ctx.fillRect(0, 0, 1200, 900);
    this.drawText(ctx, "COOL", 225, 50, "#880", 144);
    this.drawText(ctx, "24+24", 250, 100, "#808", 112);
    this.drawText(ctx, "Successive", 120, 250, "#00c", 32);
    this.drawText(ctx, "Cards", 120, 282, "#00c", 32);
    this.drawRect(ctx, 225, 245, 200, 70, "white", "#444");
    this.drawText(ctx, this.successive, 325, 250, "#44f", 64);
    this.calculateScore();
    this.drawText(ctx, "Total", 120, 330, "#c00", 32);
    this.drawText(ctx, "Score", 120, 362, "#c00", 32);
    this.drawRect(ctx, 225, 325, 200, 70, "white", "#444");
    this.drawText(ctx, this.score, 325, 330, "#f44", 64);
    this.drawCardWin(ctx);
    this.drawHandWin(ctx);
    this.drawCardList(ctx);
    this.drawRect(ctx, 460, 20, 280, 420, "white", "#ccc");
    for (let i = 0; i < numHands; i++) {
      let col = (this.handSelectable(i)) ? "#f88" : "#ccc";
      this.drawRect(ctx, 25 + i * 290, 460, 280, 420, "white", col);
    }
    if (this.top != null) {
      cardGraphics[this.top].draw(ctx, 470, 30);
    }
    for (let i = 0; i < numHands; i++) {
      this.drawHand(ctx, i);
    }
    if (this.over == 1) {
      this.drawText(ctx, "Game Over", 600, 560, "#888", 160);
      this.drawText(ctx, "Press Enter to Start", 600, 710, "#888", 64);
    } else if (this.over == 2) {
      this.drawText(ctx, "COOL24!", 600, 560, "#84f", 160);
      this.drawText(ctx, "Press Enter to Proceed", 600, 710, "#84f", 64);
    } else if (this.over == 3) {
      this.drawText(ctx, "COOL24+24!", 600, 560, "#c80", 160);
      this.drawText(ctx, "Press Enter to Start", 600, 710, "#c80", 64);
    }
  }

  // 次に配るカードを決める。ただし、現在の手札に連結するカードは、
  // そうでないカードよりも corr 倍出やすい
  selectNextCard(corr) {
    let cCards = new Array;
    let ncCards = new Array;
    for (let i = 0; i < this.rest.length; i++) {
      let c = this.rest[i];
      let con = this.connected(c, this.top);
      for (let j = 0; j < numHands; j++)
        if (this.handSelectable(j))
          con = con || this.connected(c, this.hand[j]);
      if (con)
        cCards.push(c);
      else
        ncCards.push(c);
    }
    let r = Math.random() * (cCards.length * corr + ncCards.length);
    let sel = null;
    while (r >= 0) {
      if (cCards.length > 0) {
        sel = cCards.pop();
        r -= corr;
      } else {
        sel = ncCards.pop();
        r -= 1;
      }
    }
    return sel;
  }

  // 選ばれたカードを実際に配る
  dealNextCard(pos, sel) {
    this.hand[pos] = sel;
    this.handPos[pos] = 0;
    this.rest = this.rest.filter((v) => (v != sel));
  }

  // 手札で役が出来ているかチェック（実にハードコード）
  checkHandWin() {
    let suite = new Array();
    let number = new Array();
    for (let i = 0; i < numSuites; i++)
      suite[i] = 0;
    for (let i = 0; i < numNumbers; i++)
      number[i] = 0;
    for (let i = 0; i < numHands; i++) {
      suite[Math.floor(this.hand[i] / numNumbers)] += 1;
      number[this.hand[i] % numNumbers] += 1;
    }
    for (let i = 0; i < numSuites; i++) {
      if (suite[i] == numHands) {
        if (number[0] * number[1] * number[2] * number[3] == 1 ||
            number[1] * number[2] * number[3] * number[4] == 1 ||
            number[2] * number[3] * number[4] * number[5] == 1)
          this.handWin[0] += 1;
        else
          this.handWin[2] += 1;
      }
    }
    for (let i = 0; i < numNumbers; i++)
      if (number[i] == numHands)
        this.handWin[1] += 1;
  }

  // そのカードは場に出せるか？
  handSelectable(i) {
    if (this.hand[i] == null)
      return false;
    else if (this.top == null)
      return true;
    else
      return this.connected(this.hand[i], this.top);
  }

  // カード a はカード b と連結するカードか？
  connected(a, b) {
    if (a == null || b == null)
      return false;
    let aSuite  = Math.floor(a / numNumbers);
    let aNumber = a % numNumbers;
    let bSuite  = Math.floor(b / numNumbers);
    let bNumber = b % numNumbers;
    return (aSuite == bSuite || aNumber == bNumber);
  }

  // 位置，色と文字サイズを指定して文字列を描画
  drawText(ctx, str, x, y, col, h, max = null) {
    ctx.strokeStyle = "white";
    ctx.fillStyle = col;
    ctx.font = h + 'px "Bauhaus 93"';
    ctx.textAlign = "center";
    ctx.textBaseline = "top";
    if (max) {
      ctx.fillText(str, x, y, max);
      ctx.strokeText(str, x, y, max);
    } else {
      ctx.fillText(str, x, y);
      ctx.strokeText(str, x, y);
    }
  }

  // 描画色と塗りつぶし色を指定して長方形を描画
  drawRect(ctx, x, y, w, h, col, fill = null) {
    ctx.strokeStyle = col;
    if (fill != null) {
      ctx.fillStyle = fill;
      ctx.fillRect(x, y, w, h);
    }
    ctx.strokeRect(x, y, w, h);
  }

  // 基本点のテーブルを描画
  drawCardWin(ctx) {
    this.drawText(ctx, "Cards", 810, 30, "#00c", 24);
    this.drawText(ctx, "Points",   810, 70, "#c00", 24);
    for (let i = -1; i <= 3; i++) {
      this.drawRect(ctx, 925 + i * 60, 25, 58, 38, "white", (i == 0) ? "#600" : "#444");
      this.drawRect(ctx, 925 + i * 60, 65, 58, 38, "white", (i == 0) ? "#600" : "#444");
      let cards = this.successive + i;
      if (cards >= 0 && cards <= numCards * 2) {
        this.drawText(ctx, cards, 954 + i * 60, 28, "#44f", 32, 56);
        this.drawText(ctx, cardWinTable[cards], 954 + i * 60, 68, "#f44", 32, 56);
      }
    }
  }

  // ボーナス点の獲得状況を描画
  drawHandWin(ctx) {
    for (let i = 0; i < numHandWins; i++) {
      let handWinBase = handWinTable[i] * (this.board * handWinMulti + 1);
      this.drawText(ctx, handWinName[i], 850, 127 + i * 40, "#00c", 24);
      this.drawText(ctx, handWinBase, 955, 123 + i * 40, "#00c", 32, 56);
      this.drawText(ctx, "x", 1000, 123 + i * 40, "#00c", 32);
      this.drawRect(ctx, 1015, 120 + i * 40, 58, 38, "white", (this.handWin[i] != 0) ? "#600" : "#444");
      this.drawRect(ctx, 1075, 120 + i * 40, 88, 38, "white", (this.handWin[i] != 0) ? "#600" : "#444");
      this.drawText(ctx, this.handWin[i], 1044, 123 + i * 40, "#44f", 32, 56);
      this.drawText(ctx, handWinBase * this.handWin[i], 1119, 123 + i * 40, "#f44", 32, 86);
    }
  }

  // 場に出していないカードの一覧を描画
  drawCardList(ctx) {
    this.drawRect(ctx, 800, 255, 350, 170, "white", "#444");
    for (let i = 0; i < numCards; i++) {
      if (! this.used.includes(i)) {
        let suite  = Math.floor(i / numNumbers);
        let number = i % numNumbers;
        let x = 800 + (350 * (number * 2 + 1) / (numNumbers * 2));
        let y = 260 + suite * 40 + ((this.rest.includes(i)) ? 6 : 0);
        let col = (this.rest.includes(i)) ? suiteColors[suite] : suiteColorsLite[suite];
        let size = (this.rest.includes(i)) ? 24 : 40;
        this.drawText(ctx, numberNames[number], x, y, col, size);
      }
    }
  }

  // 手札を描画
  drawHand(ctx, i) {
    if (this.hand[i] != null) {
      let x = 35 + i * 290;
      let y = 470;
      if (this.handPos[i] > 10) {
        let p = (this.handPos[i] - 10) * 0.1;
        x = (1 - p) * x + p * 470;
        y = (1 - p) * y + p * 30;
      }
      let w = (this.handPos[i] < 5) ? Math.sqrt((this.handPos[i] + 1) / 6.0) : 1;
      cardGraphics[this.hand[i]].draw(ctx, x, y, w);
    }
  }

  // ゲームをリセット（開始時に呼び出す）
  resetGame() {
    this.successive = 0;
    this.score = 0;
    this.board = 0;
    this.resetBoard();
    this.handWin = new Array();
    for (let i = 0; i < numHandWins; i++)
      this.handWin[i] = 0;
  }

  // 盤面だけリセット（開始時と、2周目開始時に呼び出す）
  resetBoard() {
    this.over = 0;
    this.used = new Array();
    this.rest = new Array();
    for (let i = 0; i< numCards; i++)
      this.rest[i] = i;
    this.hand = new Array();
    this.handPos = new Array();
    for (let i = 0; i < numHands; i++) {
      this.hand[i] = null;
      this.handPos[i] = 0;
    }
    this.top = null;
  }

  // 現在の総合得点を計算
  calculateScore() {
    this.score = cardWinTable[this.successive];
    for (let i = 0; i < numHandWins; i++)
      this.score += this.handWin[i] * handWinTable[i] * (this.board * handWinMulti + 1);
  }
}